(**********************************************************************************
 * The PEDANTIC (Proof Engine for Deductive Automation using Non-deterministic
 * Traversal of Instruction Code) verification framework
 *
 * Developed by Kenneth Roe
 * For more information, check out www.cs.jhu.edu/~roe
 *
 * Simplify.v
 * This file contains tactics for simplifying states.  Simplification is designed to
 * return equivalent states (rather than an implied state).  Hence it is applicable
 * in either a hypothesis or goal.
 *
 * Some key definitions
 *    simplifyContext
 *    simplifyExpression
 *    simplifyState'
 *    simplifyState''
 *    simplifyState'''
 *
 **********************************************************************************)

Require Export SfLib.
Require Export SfLibExtras.
Require Export ImpHeap.
Require Export AbsState.
Require Export AbsStateInstance.
Require Export Coq.Logic.FunctionalExtensionality.
Require Export Eqdep.
Require Export AbsExecute.
Require Export PickElement.

Fixpoint findMember {ev} {eq} {f} (base : @absExp ev eq f) (context : list absExp) :=
   match context with
   | (AbsFun (AbsMemberId) (element::tree))::r => if beq_absExp base element then
                                                      Some tree
                                                  else findMember base r
   | (_::r) => findMember base r
   | nil => None
   end.

(*
 * This funciton performs some contextual based simplifications of absExp's.
 *
 * Parameters:
 *    context - a list of absExp's that are known to evaluate to non-zero values
 *    state - the total absState object in which this expression was found
 *    e - the expression to simplify
 *
 * Returned:
 *    A simplified expression.
 *)
Fixpoint simplifyContext {ev} {eq} {f} {t} {ac} (context : list (@absExp ev eq f)) (state : @absState ev eq f t ac) (e : @absExp ev eq f) : absExp :=
   match e return @absExp ev eq f with
   | AbsConstVal v => AbsConstVal v
   | AbsVar v => AbsVar v
   | AbsQVar v => AbsQVar v
   | AbsFun (AbsImplyId) (p::q::nil) => match (simplifyContext context state p) with
                     | (AbsFun (AbsEqualId) ((AbsQVar l)::r::nil)) =>
                                                       if beq_absExp (subst q l r) q
                                                       then (v(l)====r)-->>(simplifyContext context state q)
                                                       else (v(l)====r)-->>(subst q l r)
                     (*| (AbsEqual r (AbsQVar l)) => match subst q l r with
                                                   | x => if beq_absExp x q then (AbsImply (AbsEqual r (AbsQVar l)) (simplifyExpression context state q))
                                                          else (AbsImply (AbsEqual r (AbsQVar l)) x)
                                                   end*)
                     | (AbsFun (AbsOrId) (p1::p2::nil)) => (
                                             (p1-->>(simplifyContext (p1::context) state q)) //\\
                                             (p2-->>(simplifyContext (p2::context) state q)))
                     | x => (x-->>(simplifyContext (x::context) state q))
                     end
   | AbsFun i l => AbsFun i (map (simplifyContext context state) l)
   end.

Fixpoint hasCell {ev} {eq} {f} {t} {ac} (state : @absState ev eq f t ac) (loc : @absExp ev eq f) :=
   match state with
   | AbsLeaf (Id 3) (*AbsCellId*) (l::v::nil) => beq_absExp l loc
   | AbsExistsT s => hasCell s loc
   | AbsAll e s => hasCell s loc
   | AbsExists e s => hasCell s loc
   | AbsEach e s => hasCell s loc
   | AbsStar l r => if (hasCell l loc:bool) then true else hasCell r loc
   | _ => false
   end.

(*
 * This funciton performs some simplifications of absExp's.
 *
 * Parameters:
 *    context - a list of absExp's that are known to evaluate to non-zero values
 *    state - the total absState object in which this expression was found
 *    e - the expression to simplify
 *
 * Returned:
 *    A simplified expression.
 *)
Fixpoint simplifyExpression {ev:Type} {eq} {f} {t} {ac} (context : list (@absExp ev eq f)) (state : @absState ev eq f t ac) (e : @absExp ev eq f) : absExp :=
   match e with
   | AbsConstVal v => AbsConstVal v
   | AbsVar v => AbsVar v
   | AbsQVar v => AbsQVar v
   | AbsFun (AbsEqualId) (e1::e2::nil) => if (beq_absExp e1 e2)
                                                 then  @AbsConstVal ev eq f (NatValue 1)
                                                 else ((@simplifyExpression ev eq f t ac context state e1)====(@simplifyExpression ev eq f t ac context state e2))
   (*| AbsFun (Id 8) (*AbsRInclude*) (element::tree::nil) => 
                     match ((simplifyExpression context state element),
                            (simplifyExpression context state tree)) with
                     | (AbsPlus base (AbsNatVal offs),t) =>
                       match findMember base context with
                       | Some tree => if ble_nat s1 offs then AbsRInclude (AbsPlus base (AbsNatVal offs)) r size fields h
                                               else if beq_absExp r1 root then AbsNatVal 1 else AbsNatVal 0
                       | None => if hasCell state (AbsPlus base (AbsNatVal offs)) then AbsNatVal 0
                                 else AbsRInclude (AbsPlus base (AbsNatVal offs)) r size fields h
                       end
                     | (e,r,h) => if hasCell state e then AbsNatVal 0 else AbsRInclude e r size fields h
                     end*)
   | AbsFun (AbsImplyId) (p::q::nil) =>
                     match (@simplifyExpression ev eq f t ac context state p) with
                     | x => (x-->>(@simplifyExpression ev eq f t ac (x::context) state q))
                     end
   | AbsFun (AbsNotId) (p::nil) => match p with
                 | AbsFun (Id 11) (*AbsAnd*) l => (AbsFun (Id 12) (*AbsOr*) (map (fun x => (~~(@simplifyExpression ev eq f t ac context state x))) l))
                 | AbsFun (Id 12) (*AbsOr*) l => (AbsFun (Id 11) (*AbsAnd*) (map (fun x => (~~(@simplifyExpression ev eq f t ac context state x))) l))
                 | AbsFun (Id 10) (*AbsNot*) (x::nil) => x
                 | _ => (~~(@simplifyExpression ev eq f t ac context state p))
                 end
   | AbsFun (AbsAndId) (p::q::nil) =>
                 match (@simplifyExpression ev eq f t ac context state p,@simplifyExpression ev eq f t ac context state q) with
                 | (#0,_) => #0
                 | (_,#0) => #0
                 | (AbsConstVal (NatValue x),q) => q
                 | (p,AbsConstVal (NatValue x)) => p
                 | (p,q) => (p //\\ q)
                 end
   | AbsFun (AbsOrId) (p::q::nil) =>
                 match (@simplifyExpression ev eq f t ac context state p,@simplifyExpression ev eq f t ac context state q) with
                 | (#0,q) => q
                 | (p,#0) => p
                 | (AbsConstVal (NatValue x),_) => (AbsConstVal (NatValue x))
                 | (_,AbsConstVal (NatValue x)) => (AbsConstVal (NatValue x))
                 | (p,q) => (p \\// q)
                 end
   | AbsFun (AbsIteId) (p::q::r::nil) => match (@simplifyExpression ev eq f t ac context state p) with
                     | AbsConstVal (NatValue x) => if beq_nat x 0
                                        then (@simplifyExpression ev eq f t ac context state r)
                                        else (@simplifyExpression ev eq f t ac context state q)
                     | x           => (AbsFun (AbsIteId) (x::(@simplifyExpression ev eq f t ac context state q)::(@simplifyExpression ev eq f t ac context state r)::nil))
                     end
   | AbsFun (AbsNthId) ((AbsFun (AbsListId) l)::(AbsConstVal x)::nil) =>
                   match x with
                   | NatValue n => nth n l (@AbsConstVal ev eq f NoValue)
                   | _ =>  (@AbsConstVal ev eq f NoValue)
                   end
   | AbsFun (AbsFindId) ((AbsFun (AbsListId) (ff::r))::f'::nil) =>
                   if beq_absExp ff f' then
                       AbsFun (AbsListId)
                       ((@simplifyExpression ev eq f t ac context state ff)::
                        (map (@simplifyExpression ev eq f t ac context state) r))
                   else
                   (AbsFun (AbsFindId)
                       ((AbsFun (AbsListId)
                            ((@simplifyExpression ev eq f t ac context state ff)::
                             (map (@simplifyExpression ev eq f t ac context state) r)))::
                        (@simplifyExpression ev eq f t ac context state f')::nil))
   | AbsFun i l => AbsFun i (map (@simplifyExpression ev eq f t ac context state) l)
   end.

Inductive substQuant {ev} {eq} {f} {t} {ac} : nat -> @absState ev eq f t ac -> @absState ev eq f t ac -> Prop :=
  | SubstQuantL : forall (s:@absState ev eq f t ac) r r' n e (s':@absState ev eq f t ac),
                  r = getRoot s ->
                  spickElement r ([v(n)====e]) r' ->
                  s' = replaceRoot s (replaceStateVar n e r') ->
                  substQuant n s s'
  | SubstQuantR : forall (s:@absState ev eq f t ac) r r' n e (s':@absState ev eq f t ac),
                  r = getRoot s ->
                  spickElement r ([e====v(n)]) r' ->
                  s' = replaceRoot s (replaceStateVar n e r') ->
                  substQuant n s s'.

Ltac solveEliminateQuant :=
    (eapply SubstQuantR;[(simpl;reflexivity)|solveSPickElement|(simpl;reflexivity)]) ||
    (eapply SubstQuantL;[(simpl;reflexivity)|solveSPickElement|(simpl;reflexivity)]).

Inductive simplifyExists {ev} {eq} {f} {t} {ac} : nat -> @absState ev eq f t ac -> @absState ev eq f t ac -> Prop :=
  | SEEliminate : forall n p e,
                  hasVnState p n=false ->
                  simplifyExists n (AbsExists e p) (removeStateVar n p)
  | SEEliminateT : forall n p,
                   hasVnState p n=false ->
                   simplifyExists n (AbsExistsT p) (removeStateVar n p)
  | SEEliminateQuantT : forall n p p',
                        substQuant n p p' ->
                        simplifyExists n (AbsExistsT p) (AbsExistsT p')
  | SEEliminateQuant : forall n p p' e,
                       substQuant n p p' ->
                       simplifyExists n (AbsExists e p) (AbsExists e p').
  (*| SESplitNat : forall p p1 p2,
                 separateV0components p p1 p2 ->
                 ~(empty_state p2) ->
                 simplifyExists (AbsExists AbsNat p) ((AbsExists AbsNat p1) ** (instantiateNatState nil p2 0))*)
  (*| SESplitHeap : forall p p1 p2,
                  separateV0components p p1 p2 ->
                  ~(empty_state p2) ->
                  simplifyExists (AbsExists AbsHeap p) ((AbsExists AbsHeap p1) ** (instantiateHeapState nil p2 empty_heap)).*)

Ltac solveSimplifyExists :=
    solve [(eapply SEEliminate;simpl;reflexivity) |
           (eapply SEEliminateT;simpl;reflexivity) |
           (eapply SEEliminateQuantT;solveEliminateQuant) |
           (eapply SEEliminateQuant;solveEliminateQuant) ].

Ltac solveSimplifyExistsPPP :=
    solve [(eapply SEEliminate;simpl;reflexivity) |
           (eapply SEEliminateT;simpl;reflexivity) ].

Fixpoint reduceFieldRef {ev} {eq} {f} (n : nat ) (e : @absExp ev eq f) : option (absExp * list absExp) :=
    match e with
    | AbsFun (AbsFindId) ((AbsFun (AbsListId) (f::l))::(AbsQVar x)::nil) =>
          if beq_nat x n then
              Some (AbsFun (AbsFindId) ((AbsQVar (S n))::(AbsQVar x)::nil),l)
          else
              None
    | AbsFun i l => match (fix go l := match l with
                                       | (ff::r) => match @reduceFieldRef ev eq f n ff with
                                                   | Some (a,b) => Some ((a::r),b)
                                                   | None => match go r with
                                                             | Some (a,b) => Some (ff::a,b)
                                                             | None => None
                                                             end
                                                   end
                                       | nil => None
                                       end) l with
                    | Some (a,b) => Some (AbsFun i a,b)
                    | None => None
                    end
    | x => None
    end.

Inductive allPresent {ev} {eq} {f} {t} {ac} : list (@absExp ev eq f) -> (@absState ev eq f t ac) -> Prop :=
  | APNil : forall s, allPresent nil s
  | APCell : forall s s' ff r l,
             spickElement s (l |-> ff) s' ->
             allPresent r s ->
             allPresent (ff::r) s
  | APVar :  forall v r s,
             allPresent r s ->
             allPresent ((AbsVar v)::r) s
  | APTree : forall s s' ff size fields root r,
             spickElement s (TREE(root,ff,size,fields)) s' ->
             allPresent r s ->
             allPresent (ff::r) s.

Ltac allPresent :=
    solve [ eapply APNil | eapply APVar;allPresent | eapply APCell;[solveSPickElement|allPresent] | eapply APTree;[solveSPickElement|allPresent]].


(*
 * This inductive defintion is used to simplify states.  There are three variants of this definition
 * due to performance tuning.
 *
 * Parameters:
 *    #1 - current deBruin quantifier variable level
 *    #2 - entire state
 *    #3 - subterm currently being simplified
 *    #4 - result of simplification
 *)
Inductive simplifyState' {ev} {eq} {f} {t} {ac} : nat -> @absState ev eq f t ac -> @absState ev eq f t ac -> @absState ev eq f t ac -> Prop :=
  | SETSimplifyInExists : forall p p' n x m,
                          simplifyState' (S n) x p p' ->
                          simplifyState' n x (AbsExists m p) (AbsExists m p')
  | SETSimplifyInExistsT : forall p p' n x,
                           simplifyState' (S n) x p p' ->
                           simplifyState' n x (AbsExistsT p) (AbsExistsT p')
  | SETDumpAll :           forall x s s' s'' l v n p,
                           spickElement s (l |-> v) s' ->
                           spickElement s (AbsAll (TreeRecords(v)) p) s'' ->
                           simplifyState' n x s s''
  | SETDumpAll2 :           forall x s s'' v n p,
                            spickElement s (AbsAll (TreeRecords(AbsVar v)) p) s'' ->
                            simplifyState' n x s s''
  | SETReduceFindRecord: forall n x s s' s'' p p' p'' v l root shape s''',
                         spickElement s (AbsAll (TreeRecords(v)) ([p])) s' ->
                         Some (p',l) = reduceFieldRef n p ->
                         allPresent l s ->
                         spickElement s (AbsLeaf (AbsTreeId) (root::v::shape)) s'' ->
                         In v l ->
                         p'' = replaceExpVar (n+1) v p' ->
                         s''' = AbsStar s' (AbsAll (TreeRecords(v)) ([p''])) ->
                         simplifyState' n x s s'''
  | SETResolvePredicates1 : forall x s s' s'' p1 p2 n,
                            spickElement s ([p1 \\// p2]) s' ->
                            spickElement s' ([~~p1]) s'' ->
                            simplifyState' n x s ([p2] ** s')
  | SETResolvePredicates2 : forall x s s' s'' p1 p2 n,
                            spickElement s ([p1 \\// p2]) s' ->
                            spickElement s' ([~~p2]) s'' ->
                            simplifyState' n x s ([p1] ** s')
  | SETResolvePredicates3 : forall x s s' s'' p1 p2 n,
                            spickElement s ([~~p1 \\// p2]) s' ->
                            spickElement s' ([p1]) s'' ->
                            simplifyState' n x s ([p2] ** s')
  | SETResolvePredicates4 : forall x s s' s'' p1 p2 n,
                            spickElement s ([p1 \\// ~~p2]) s' ->
                            spickElement s' ([p2]) s'' ->
                            simplifyState' n x s ([p1] ** s')
  | SETChooseVariable1 : forall p p' p'' n x v v',
                        spickElement p ([(AbsVar (Id v))====(AbsVar (Id v'))]) p' ->
                        v < v' ->
                        p'' = substVarState p' (Id v') (AbsVar (Id v)) ->
                        p''<> p' ->
                        simplifyState' n x p (AbsStar ([(AbsVar (Id v))====(AbsVar (Id v'))]) p'')
  | SETChooseVariable2 : forall p p' p'' n x v v',
                        spickElement p ([(AbsVar (Id v'))====(AbsVar (Id v))]) p' ->
                        v < v' ->
                        p'' = substVarState p' (Id v') (AbsVar (Id v)) ->
                        simplifyState' n x p (AbsStar ([(AbsVar (Id v))====(AbsVar (Id v'))]) p'')
  | SETChooseConstant1 : forall p p' p'' n x v c,
                        spickElement p ([(AbsConstVal c)====(AbsVar (Id v))]) p' ->
                        p'' = substVarState p' (Id v) (AbsConstVal c) ->
                        p'' <> p' ->
                        simplifyState' n x p (AbsStar ([(AbsConstVal c)====(AbsVar (Id v))]) p'')
  | SETChooseConstant2 : forall p p' p'' n x v c,
                        spickElement p ([(AbsVar (Id v))====(AbsConstVal c)]) p' ->
                        p'' = substVarState p' (Id v) (AbsConstVal c) ->
                        p'' <> p' ->
                        simplifyState' n x p (AbsStar ([(AbsVar (Id v))====(AbsConstVal c)]) p'')
  | SETQVarElim1 : forall p p' p'' n x v c,
                   spickElement p ([(AbsConstVal c)====(AbsQVar v)]) p' ->
                   p'' = replaceStateVar v (AbsConstVal c) p' ->
                   simplifyState' n x p p''
  | SETQVarElim2 : forall p p' p'' n x v c,
                   spickElement p ([(AbsQVar v)====(AbsConstVal c)]) p' ->
                   p'' = replaceStateVar v (AbsConstVal c) p' ->
                   simplifyState' n x p p''
  | SETQVarElim3 : forall p p' p'' n x v v',
                   spickElement p ([(AbsVar v')====(AbsQVar v)]) p' ->
                   p'' = replaceStateVar v (AbsVar v') p' ->
                   simplifyState' n x p p''
  | SETQVarElim4 : forall p p' p'' n x v v',
                   spickElement p ([(AbsQVar v)====(AbsVar v')]) p' ->
                   p'' = replaceStateVar v (AbsVar v') p' ->
                   simplifyState' n x p p''
  | SETQVarElim5 : forall p p' p'' n x v i l,
                   spickElement p ([(AbsFun i l)====(AbsQVar v)]) p' ->
                   p'' = replaceStateVar v (AbsFun i l) p' ->
                   simplifyState' n x p p''
  | SETQVarElim6 : forall p p' p'' n x v i l,
                   spickElement p ([(AbsQVar v)====(AbsFun i l)]) p' ->
                   p'' = replaceStateVar v (AbsFun i l) p' ->
                   simplifyState' n x p p''
  | SETFunElim1 : forall p p' p'' n x v i l,
                  spickElement p ([(AbsVar v)====(AbsFun i l)]) p' ->
                  p'' = replaceStateExp (AbsFun i l) (AbsVar v) p' ->
                  p'' <> p' ->
                  simplifyState' n x p (AbsStar ([(AbsFun i l)====(AbsVar v)]) p'')
  | SETFunElim2 : forall p p' p'' n x v i l,
                  spickElement p ([(AbsFun i l)====(AbsVar v)]) p' ->
                  p'' = replaceStateExp (AbsFun i l) (AbsVar v) p' ->
                  p'' <> p' ->
                  simplifyState' n x p (AbsStar ([(AbsFun i l)====(AbsVar v)]) p'').

(*
 * This inductive defintion is used to simplify states
 *
 * Parameters:
 *    #1 - current deBruin quantifier variable level
 *    #2 - entire state
 *    #3 - subterm currently being simplified
 *    #4 - result of simplification
 *)
Inductive simplifyState'' {ev} {eq} {f} {t} {ac} : nat -> @absState ev eq f t ac -> @absState ev eq f t ac -> @absState ev eq f t ac -> Prop :=
  | SEComponent1 : forall n p1 p1' p2 x,
                   simplifyState'' n x p1 p1' ->
                   simplifyState'' n x (p1 ** p2) (p1' ** p2)
  | SEComponent2 : forall p1 p2 p2' n x,
                   simplifyState'' n x p2 p2' ->
                   simplifyState'' n x (p1 ** p2) (p1 ** p2')
  | SESimplifyEmpty1 : forall n p x,
                       simplifyState'' n x (AbsEmpty ** p) p
  | SESimplifyEmpty2 : forall n p x,
                       simplifyState'' n x (p ** AbsEmpty) p
  | SESimplifyPred : forall n x xx,
                      xx <> 0 ->
                      simplifyState'' n x ([#xx]) AbsEmpty
  | SESimplifyInExists: forall p p' n e x,
                   simplifyState'' (S n) x p p' ->
                   simplifyState'' n x (AbsExists e p) (AbsExists e p')
  | SESimplifyInExistsT: forall p p' n x,
                   simplifyState'' (S n) x p p' ->
                   simplifyState'' n x (AbsExistsT p) (AbsExistsT p')
  | SESimplifyInAll: forall p p' n e x,
                   simplifyState'' (S n) x p p' ->
                   simplifyState'' n x (AbsAll e p) (AbsAll e p')
  | SESimplifyInEach: forall p p' n e x,
                      simplifyState'' (S n) x p p' ->
                      simplifyState'' n x (AbsEach e p) (AbsEach e p')
  | SEPropLeft : forall n p1 p2 e xx,
                 simplifyState'' n xx ((AbsExists e p1) ** p2) (AbsExists e (p1 ** (addStateVar n p2)))
  | SEPropRight : forall n p1 p2 e xx,
                 simplifyState'' n xx (p1 ** (AbsExists e p2)) (AbsExists e ((addStateVar n p1) ** p2))
  | SEPropLeftT : forall n p1 p2 xx, 
                  simplifyState'' n xx ((AbsExistsT p1) ** p2) (AbsExistsT (p1 ** (addStateVar n p2)))
  | SEPropRightT : forall n p1 p2 xx,
                   simplifyState'' n xx (p1 ** (AbsExistsT p2)) (AbsExistsT ((addStateVar n p1) ** p2))
  | SESimplifyExists: forall p p' n x e,
                      simplifyExists n (AbsExists e p) p' ->
                      simplifyState'' n x (AbsExists e p) p'
  | SESimplifyExistsT: forall p p' n x,
                      simplifyExists n (AbsExistsT p) p' ->
                      simplifyState'' n x (AbsExistsT p) p'
  | SESimplifyPredicate : forall i p p' x n,
                          p' = map (fun xx => simplifyContext nil x (@simplifyExpression ev eq f t ac nil x xx)) p ->
                          p' <> p ->
                          simplifyState'' n x (AbsLeaf i p) (AbsLeaf i p')
  | SEAllDomSplit : forall p1 p2 l x n ff,
                    p2 = fold_left (fun x y => (AbsStar x y))
                                   (map (fun x => (AbsAll (TreeRecords(x)) p1)) l)
                                   (removeStateVar n (replaceStateVar n ff p1)) ->
                 simplifyState'' n x (AbsAll (TreeRecords(AbsFun (AbsListId) (ff::l))) p1) p2
  | SEAllSplit : forall p1 p2 e x n,
                 simplifyState'' n x (AbsAll e (p1 ** p2)) ((AbsAll e p1) ** (AbsAll e p2))
  | SEAllElim1 : forall n p e x ee,
                 (hasVnExp p n=false) ->
                 (hasVnExp e n=false) ->
                 simplifyState'' n x (AbsAll ee ([(v(n)====e)-->>p])) ([p])
  | SEPredicateAndSplit : forall p1 p2 x n,
                 simplifyState'' n x ([p1 //\\ p2]) (([p1]) ** ([p2]))
  | SESimplifySplitPredicate : forall p p' x n,
                               p' = @simplifyExpression ev eq f t ac nil x p ->
                               p' <> p ->
                               simplifyState'' n x ([p]) ([p'])
  | SESSimplifyIdentity : forall n p x, simplifyState'' n x p p.


(*
 * This inductive defintion is used to simplify states
 *
 * Parameters:
 *    #1 - current deBruin quantifier variable level
 *    #2 - entire state
 *    #3 - subterm currently being simplified
 *    #4 - result of simplification
 *)
Inductive simplifyState''' {ev} {eq} {f} {t} {ac} : nat -> @absState ev eq f t ac -> @absState ev eq f t ac -> @absState ev eq f t ac -> Prop :=
  | SEComponent1' : forall n p1 p1' p2 x,
                   simplifyState''' n x p1 p1' ->
                   simplifyState''' n x (p1 ** p2) (p1' ** p2)
  | SEComponent2' : forall p1 p2 p2' n x,
                   simplifyState''' n x p2 p2' ->
                   simplifyState''' n x (p1 ** p2) (p1 ** p2')
  | SESimplifyEmpty1' : forall n p x,
                       simplifyState''' n x (AbsEmpty ** p) p
  | SESimplifyEmpty2' : forall n p x,
                       simplifyState''' n x (p ** AbsEmpty) p
  | SESimplifyPred' : forall n x xx,
                      xx <> 0 ->
                      simplifyState''' n x ([#xx]) AbsEmpty
  | SESimplifyInExists': forall p p' n e x,
                   simplifyState''' (S n) x p p' ->
                   simplifyState''' n x (AbsExists e p) (AbsExists e p')
  | SESimplifyInExistsT': forall p p' n x,
                   simplifyState''' (S n) x p p' ->
                   simplifyState''' n x (AbsExistsT p) (AbsExistsT p')
  | SESimplifyInAll': forall p p' n e x,
                   simplifyState''' (S n) x p p' ->
                   simplifyState''' n x (AbsAll e p) (AbsAll e p')
  | SESimplifyInEach': forall p p' n e x,
                   simplifyState''' (S n) x p p' ->
                   simplifyState''' n x (AbsEach e p) (AbsEach e p')
  | SEPropLeft' : forall n p1 p2 e xx,
                 simplifyState''' n xx ((AbsExists e p1) ** p2) (AbsExists e (p1 ** (addStateVar n p2)))
  | SEPropRight' : forall n p1 p2 e xx,
                 simplifyState''' n xx (p1 ** (AbsExists e p2)) (AbsExists e ((addStateVar n p1) ** p2))
  | SEPropLeftT' : forall n p1 p2 xx,
                 simplifyState''' n xx ((AbsExistsT p1) ** p2) (AbsExistsT (p1 ** (addStateVar n p2)))
  | SEPropRightT' : forall n p1 p2 xx,
                 simplifyState''' n xx (p1 ** (AbsExistsT p2)) (AbsExistsT ((addStateVar n p1) ** p2))
  | SESimplifyExists': forall p p' n x e,
                        simplifyExists n (AbsExists e p) p' ->
                        simplifyState''' n x (AbsExists e p) p'
  | SESimplifyExistsT': forall p p' n x,
                        simplifyExists n (AbsExistsT p) p' ->
                        simplifyState''' n x (AbsExistsT p) p'
  | SESimplifyPredicate' : forall p p' x n,
                          p' = simplifyContext nil x (@simplifyExpression ev eq f t ac nil x p) ->
                          p' <> p ->
                          simplifyState''' n x ([p]) ([p'])
  | SEAllSplit' : forall p1 p2 x n e,
                 simplifyState''' n x (AbsAll e (p1 ** p2)) ((AbsAll e p1) ** (AbsAll e p2))
  | SEAllElim1' : forall n p ee x e,
                 (hasVnExp p n=false) ->
                 (hasVnExp e n=false) ->
                 simplifyState''' n x (AbsAll ee ([(v(n)====e)-->>p])) ([p])
  | SEPredicateAndSplit' : forall p1 p2 x n,
                 simplifyState''' n x ([p1 //\\ p2]) (([p1]) ** ([p2]))
  | SESimplifySplitPredicate' : forall p p' x n,
                               p' = @simplifyExpression ev eq f t ac nil x p ->
                               p' <> p ->
                               simplifyState''' n x ([p]) ([p'])
  | SESSimplifyIdentity' : forall n p x, simplifyState''' n x p p.

Inductive simplifyState {ev} {eq} {f} {t} {ac} : nat -> @absState ev eq f t ac -> @absState ev eq f t ac -> @absState ev eq f t ac -> Prop :=
  | SEP : forall n x p p', @simplifyState' ev eq f t ac n x p p' -> simplifyState n x p p'
  | SEPP : forall n x p p', @simplifyState'' ev eq f t ac n x p p' -> simplifyState n x p p'.


Ltac solveOr := solve [(left;reflexivity) | (right;solveOr)].

Ltac simplifyStateP :=
    solve [(eapply SETResolvePredicates1;[(simpl;reflexivity)|solveSPickElement|solveSPickElement|(simpl;reflexivity)]) |
           (eapply SETDumpAll;[solveSPickElement|solveSPickElement]) |
           (eapply SETDumpAll2;[solveSPickElement]) |
           (eapply SETResolvePredicates1;[(simpl;reflexivity)|solveSPickElement|solveSPickElement|(simpl;reflexivity)]) |
           (eapply SETResolvePredicates2;[(simpl;reflexivity)|solveSPickElement|solveSPickElement|(simpl;reflexivity)]) |
           (eapply SETResolvePredicates3;[(simpl;reflexivity)|solveSPickElement|solveSPickElement|(simpl;reflexivity)]) |
           (eapply SETResolvePredicates4;[(simpl;reflexivity)|solveSPickElement|solveSPickElement|(simpl;reflexivity)]) |
           (eapply SETReduceFindRecord;[solveSPickElement|(simpl;reflexivity)|allPresent|solveSPickElement|(simpl;solveOr)|(simpl;reflexivity)|(simpl;reflexivity)]) |
           (eapply SETSimplifyInExists;simplifyStateP) |
           (eapply SETSimplifyInExistsT;simplifyStateP) |
           (eapply SETChooseConstant1;[solveSPickElement|crunch|(intro X;inversion X)]) |
           (eapply SETChooseConstant2;[solveSPickElement|crunch|(intro X;inversion X)]) |
           (eapply SETFunElim1;[solveSPickElement|crunch|(intro X;inversion X)]) |
           (eapply SETFunElim2;[solveSPickElement|crunch|(intro X;inversion X)]) |
           (eapply SETChooseVariable1;[solveSPickElement|crunch|crunch|(intro X;inversion X)]) |
           (eapply SETChooseVariable2;[solveSPickElement|crunch|crunch])
           (*(eapply SETQVarElim1;[solveSPickElement|crunch]) |
           (eapply SETQVarElim2;[solveSPickElement|crunch]) |
           (eapply SETQVarElim3;[solveSPickElement|crunch]) |
           (eapply SETQVarElim4;[solveSPickElement|crunch]) |
           (eapply SETQVarElim5;[solveSPickElement|crunch]) |
           (eapply SETQVarElim6;[solveSPickElement|crunch])*)
           ].

Ltac simplifyStatePP :=
    solve [(eapply SESimplifyEmpty1) |
           (eapply SESimplifyEmpty2) |
           (eapply SESimplifyPred;omega) |
           (eapply SEAllElim1;simpl;reflexivity) |
           (eapply SESimplifyPredicate;[(simpl;reflexivity)|(let X := fresh in intro X;inversion X)]) |
           (eapply SESimplifyExists;solveSimplifyExists) |
           (eapply SESimplifyExistsT;solveSimplifyExists) |
           eapply SEPredicateAndSplit |
           eapply SEAllSplit |
           (eapply SEAllDomSplit;simpl;reflexivity) |
           eapply SEPropLeft |
           eapply SEPropRight |
           eapply SEPropLeftT |
           eapply SEPropRightT |
           (eapply SESimplifyInExists;simplifyStatePP) |
           (eapply SESimplifyInExistsT;simplifyStatePP) |
           (eapply SESimplifyInAll;simplifyStatePP) |
           (eapply SESimplifyInEach;simplifyStatePP) |
           (eapply SEComponent1;simplifyStatePP) |
           (eapply SEComponent2;simplifyStatePP)
           ].

Ltac simplifyStatePPP :=
    solve [(eapply SESimplifyEmpty1') |
           (eapply SESimplifyEmpty2') |
           (eapply SEAllElim1';simpl;reflexivity) |
           (eapply SESimplifyPred';omega) |
           (eapply SESimplifyPredicate';[(simpl;reflexivity)|(let X := fresh in intro X;inversion X)]) |
           (eapply SESimplifyExists';solveSimplifyExistsPPP) |
           (eapply SESimplifyExistsT';solveSimplifyExistsPPP) |
           eapply SEPredicateAndSplit' |
           eapply SEAllSplit' |
           eapply SEPropLeft' |
           eapply SEPropRight' |
           (eapply SESimplifyInExists';simplifyStatePPP) |
           (eapply SESimplifyInExistsT';simplifyStatePPP) |
           (eapply SESimplifyInAll';simplifyStatePPP) |
           (eapply SESimplifyInEach';simplifyStatePPP) |
           (eapply SEComponent1';simplifyStatePPP) |
           (eapply SEComponent2';simplifyStatePPP)
           ].

(*
 * All of the theorems and tactics below are used to apply simplification in different
 * situates in a program proof.
 *)
Ltac simplifyState :=
    solve [(eapply SEP;simplifyStateP) | (eapply SEPP;simplifyStatePP)].

Theorem simplifyEquiv1 {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' state,
    @simplifyState ev eq f t ac 0 s s s' ->
    realizeState s nil state ->
    realizeState s' nil state.
Proof. admit. Qed.

Theorem simplifyEquiv' {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' state,
    @simplifyState' ev eq f t ac 0 s s s' ->
    realizeState s nil state ->
    realizeState s' nil state.
Proof. admit. Qed.

Theorem simplifyEquiv'' {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' state,
    @simplifyState'' ev eq f t ac 0 s s s' ->
    realizeState s nil state ->
    realizeState s' nil state.
Proof. admit. Qed.

Ltac simplifyHyp X:=
    repeat (eapply simplifyEquiv1 in X;[simpl in X|simplifyState]).

Theorem simplifyEquiv2 {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' state,
    @simplifyState ev eq f t ac 0 s s s' ->
    realizeState s' nil state -> realizeState s nil state.
Proof. admit. Qed.

Theorem simplifyPre {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' cmd post r,
    @simplifyState ev eq f t ac 0 s s s' ->
    {{ s' }} cmd {{ post, r }} ->
    {{ s }} cmd {{ post, r }}.
Proof. admit. Qed.

Theorem tsimplifyPre {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' cmd post r,
    {{ s' }} cmd {{ post, r }} ->
    @simplifyState ev eq f t ac 0 s s s' ->
    {{ s }} cmd {{ post, r }}.
Proof. admit. Qed.

Ltac simp := eapply simplifyPre;[simplifyState|(instantiate;simpl)] || idtac.
Ltac simplifyPre := repeat (progress simp).

Ltac simpBasic := eapply (@simplifyPre unit eq_unit
                    (@basicEval unit)
                    (@basicState unit) (@basicAccumulate unit eq_unit) tt)
                    ;[simplifyState|(instantiate;simpl)] || idtac.
Ltac simplifyPreBasic := repeat (progress simpBasic).

Theorem simplifyPre' {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' cmd post r,
    @simplifyState' ev eq f t ac 0 s s s' ->
    {{ s' }} cmd {{ post, r }} ->
    {{ s }} cmd {{ post, r }}.
Proof. admit. Qed.

Theorem simplifyPre'' {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' cmd post r,
    @simplifyState'' ev eq f t ac 0 s s s' ->
    {{ s' }} cmd {{ post, r }} ->
    {{ s }} cmd {{ post, r }}.
Proof. admit. Qed.

Theorem simplifyPre''' {ev} {eq} {f} {t} {ac} : forall (s : @absState ev eq f t ac) s' cmd post r,
    @simplifyState''' ev eq f t ac 0 s s s' ->
    {{ s' }} cmd {{ post, r }} ->
    {{ s }} cmd {{ post, r }}.
Proof. admit. Qed.

Theorem mergeSimplifyLeft {ev} {eq} {f} {t} {ac} : forall (P1 : @absState ev eq f t ac) P1' P2 P,
    @simplifyState ev eq f t ac 0 P1 P1 P1' ->
    mergeStates P1' P2 P ->
    mergeStates P1 P2 P.
Proof. admit. Qed.

Theorem mergeSimplifyRight {ev} {eq} {f} {t} {ac} : forall (P1 : @absState ev eq f t ac) P2 P2' P,
    @simplifyState ev eq f t ac 0 P2 P2 P2' ->
    mergeStates P1 P2' P ->
    mergeStates P1 P2 P.
Proof. admit. Qed.

Ltac simpleft := eapply mergeSimplifyLeft;[simplifyState|(instantiate;simpl)] || idtac.
Ltac mergeSimplifyLeft := repeat (progress simpleft).

Ltac simpright := eapply mergeSimplifyRight;[simplifyState|(instantiate;simpl)] || idtac.
Ltac mergeSimplifyRight := repeat (progress simpright).
