(**********************************************************************************
 * The PEDANTIC (Proof Engine for Deductive Automation using Non-deterministic
 * Traversal of Instruction Code) verification framework
 *
 * Developed by Kenneth Roe
 * For more information, check out www.cs.jhu.edu/~roe
 *
 * StateImplication.v
 * This file contains many defintions and tactics useful in proving that one state
 * implies another.  The key tactic is one that pairs off components of a state.
 *
 * Key definitions:
 *     stateImplication
 *     prove_implication
 *
 **********************************************************************************)

Require Export SfLib.
Require Export ImpHeap.
Require Export AbsState.
Require Export PickElement.
Require Export AbsStateInstance.

(* Glue stuff *)
Fixpoint no_cell_terms {ev} {eq} {f} {t} {ac} (s : @absState ev eq f t ac) : bool :=
    match s with
    | AbsStar a b => if no_cell_terms a then no_cell_terms b else false
    | AbsExists e s => no_cell_terms s
    | AbsExistsT s => no_cell_terms s
    | AbsAll e s => no_cell_terms s
    | AbsEach e s => no_cell_terms s
    | AbsLeaf (AbsCellId) l => false
    | _ => true
    end.

Fixpoint no_r_terms {ev} {eq} {f} {t} {ac} (s : @absState ev eq f t ac) : bool :=
    match s with
    | AbsStar a b => if no_r_terms a then no_r_terms b else false
    | AbsExists e s => no_r_terms s
    | AbsExistsT s => no_r_terms s
    | AbsAll e s => no_r_terms s
    | AbsEach e s => no_r_terms s
    | AbsLeaf (Id x) l => if (beq_nat x 3) (*AbsCell*) then true else if (beq_nat x 1) (*AbsPredicate*) then true else false
    | _ => true
    end.

Fixpoint r_term_list {ev} {eq} {f} {t} {ac} (s : @absState ev eq f t ac) : list absState :=
    match s return list (@absState ev eq f t ac) with
    | AbsStar a b => (r_term_list a)++(r_term_list b)
    | AbsExists e s => r_term_list s
    | AbsExistsT s => r_term_list s
    | AbsAll e s => r_term_list s
    | AbsEach e s => r_term_list s
    | AbsLeaf (Id x) l => if beq_nat x 1 then nil else if beq_nat x 3 then nil else (AbsLeaf (Id x) l)::nil
    | _ => nil
    end.

Fixpoint predicate_list {ev} {eq} {f} {t} {ac} (s : @absState ev eq f t ac) : list absState :=
    match s return list (@absState ev eq f t ac) with
    | AbsStar a b => (predicate_list a)++(predicate_list b)
    | AbsExists e s => predicate_list s
    | AbsExistsT s => predicate_list s
    | AbsAll e s => predicate_list s
    | AbsEach e s => predicate_list s
    | AbsLeaf (AbsPredicateId) l => (AbsLeaf (AbsPredicateId) l)::nil
    | _ => nil
    end.

Fixpoint strip_front_exists {ev} {eq} {f} {t} {ac} (s : @absState ev eq f t ac) : (absState * nat) :=
    match s with
    | AbsExistsT s => match strip_front_exists s with
                      (st,n) => (st,n+1)
                      end
    (*| AbsExists i l s => match strip_front_exists s with
                         (st,n) => (st,n+1)
                         end*)
    | _ => (s,0)
    end.

Fixpoint map_over (l : list (nat * nat)) (v : nat) :=
    match l with
    | nil => 0
    | ((x1,x2)::r) => if beq_nat x2 v then x1 else map_over r v
    end.

Fixpoint map_over_exp {ev} {eq} {f} (l : list (nat * nat)) (limit1 : nat) (limit2 : nat) (e : @absExp ev eq f) : absExp :=
   match e return @absExp ev eq f with
   | AbsConstVal v => AbsConstVal v
   | AbsVar v => AbsVar v
   | AbsQVar v => AbsQVar (if ble_nat limit2 v then (v+limit1)-limit2 else map_over l v)
   | AbsFun i ll => AbsFun i (map (map_over_exp l limit1 limit2) ll)
   end.

Fixpoint map_over_state {ev} {eq} {f} {t} {ac} (l : list (nat * nat)) (limit1 : nat) (limit2 : nat) (s : @absState ev eq f t ac) : absState :=
   match s return @absState ev eq f t ac with
    | AbsStar s1 s2 => (AbsStar (map_over_state l limit1 limit2 s1) (map_over_state l limit1 limit2 s2))
    | AbsExists e s => AbsExists (map_over_exp l limit1 limit2 e) (map_over_state l limit1 limit2 s)
    | AbsExistsT s => AbsExistsT (map_over_state l limit1 limit2 s)
    | AbsAll e s => AbsAll (map_over_exp l limit1 limit2 e) (map_over_state l limit1 limit2 s)
    | AbsEach e s => AbsEach (map_over_exp l limit1 limit2 e) (map_over_state l limit1 limit2 s)
    | AbsEmpty => AbsEmpty
    | AbsLeaf i ll => AbsLeaf i (map (map_over_exp l limit1 limit2) ll)
    | AbsAccumulate i l1 i2 a l2 r => AbsAccumulate i (map (map_over_exp l limit1 limit2) l1) i2 (map_over_exp l limit1 limit2 a) (map (map_over_exp l limit1 limit2) l2) (map_over_exp l limit1 limit2 r)
   end.

Fixpoint prove_state_implication {ev} {eq} {f} {t} {ac} (tl : nat) (s1: @absState ev eq f t ac) (s2 : @absState ev eq f t ac) (e : env) (h : heap) : Prop :=
    match tl return Prop with
    | 0 => realizeState s1 nil (e,h) -> realizeState s2 nil (e,empty_heap)
    | S n => exists x, prove_state_implication n (instantiateState s1 x) (instantiateState s2 x) e h
    end.

(*
 * This top level definition is responsible for proving implications.  It works by first pairing off
 * identical components and then setting up a proof goal for the remainder.
 *)
Inductive prove_implication {ev} {eq} {f} {t} {ac} : list (nat * nat) -> @absState ev eq f t ac -> nat -> @absState ev eq f t ac -> nat -> list (nat * nat) -> @absState ev eq f t ac -> Prop :=
    | CILPairR : forall s1 s2 s1' s2' l1 l2 vars vars' vars'' limit1 limit2 tl,
               pick2RsNi s1 s2 vars limit1 limit2 nil nil l1 l2 s1' s2' vars' ->
               prove_implication vars' s1' limit1 s2' limit2 vars'' tl ->
               prove_implication vars s1 limit1 s2 limit2 vars'' tl
    (*| CILPairCell : forall s1 s2 s1' s2' loc1 loc2 val1 val2 vars vars' vars'' limit1 limit2 tl,
               pick2Cells s1 s2 vars limit1 limit2 nil nil loc1 loc2 val1 val2 s1' s2' vars' ->
               prove_implication vars' s1' limit1 s2' limit2 vars'' tl ->
               prove_implication vars s1 limit1 s2 limit2 vars'' tl*)
    | CILFinish : forall vars limit1 limit2 s1 s2,
                  length vars = limit2 ->
                  prove_implication vars s1 limit1 s2 limit2 vars s1.

Ltac prove_implication := (eapply CILPairR;[solvePick2RsNi | prove_implication]) ||
                          (eapply CILFinish;simpl;reflexivity).

(*
 * The top level state implication theorem.  One state implies another if we
 * can first pair off many of the identical components and then prove that
 * the first state implies the remaining components.
 *)
Theorem stateImplication {ev} {eq} {f} {t} {ac} : forall s state1 (state2 : @absState ev eq f t ac) state1' state1'' tl1 tl2 state2' state2'' vars,
    realizeState state1 nil s ->
    (state1',tl1) = remove_top_existentials state1 ->
    (state2',tl2) = remove_top_existentials state2 ->
    prove_implication nil state2' tl2 state1' tl1 vars state2'' ->
    state1'' = map_over_state vars tl1 tl2 state1' ->
    (forall e h, prove_state_implication tl2 state1'' state2'' e h) ->
    realizeState state2 nil s.
Proof. admit. Qed.

(*
 * The tactics below are all useful in applying the stateImplication
 * theorem above and performing other useful tasks in proving an
 * implication
 *)
Ltac stateImplication :=
    eapply stateImplication;[
        crunch |
        (simpl; reflexivity) |
        (simpl; reflexivity) |
        prove_implication |
        (simpl; reflexivity) |
        idtac].

Ltac reduceHyp :=
    match goal with
    | [ H: 1 = 0 |- _ ] => inversion H
    | [ H: 0 = 1 |- _ ] => inversion H
    | [ H: true = false |- _ ] => inversion H
    | [ H: false = true |- _ ] => inversion H
    | [ H: None = Some _ |- _ ] => inversion H
    | [ H: Some _ = None |- _ ] => inversion H
    | [ H: 0 <> 0 |- _ ] => omega
    | [ H: 1 <> 1 |- _ ] => omega
    | [H: context [if beq_nat ?X ?Y then _ else _] |- _] => let x:= fresh in remember (beq_nat X Y) as x; destruct x; simpl in H
    | [ H: true = beq_nat _ _ |- _ ] => apply beq_nat_eq in H;subst
    (*| [H: absEval (AbsVar _) _ _ |- _] => inversion H;subst;clear H
    | [H: absEval (AbsNatVal _) _ _ |- _] => inversion H;subst;clear H
    | [H: absEval (AbsHeapVal _) _ _ |- _] => inversion H;subst;clear H
    | [H: context [beq_nat 1 1] |- _ ] => simpl in H
    | [H: context [beq_nat 1 0] |- _ ] => simpl in H
    | [H: context [beq_nat 0 1] |- _ ] => simpl in H
    | [H: context [beq_nat 0 0] |- _ ] => simpl in H*)
    | [H: 1 = 1 |- _] => inversion H; subst; clear H
    | [H: 1 = 0 |- _] => inversion H; subst; clear H
    | [H: 0 = 1 |- _] => inversion H; subst; clear H
    | [H: 0 = 0 |- _] => inversion H; subst; clear H
    | [H: NatValue _ = NatValue _ |- _] => inversion H; subst; clear H
    (*| [H: HeapValue _ = HeapValue _ |- _] => inversion H; subst; clear H*)
    | [H: OtherValue _ = OtherValue _ |- _] => inversion H; subst; clear H
    | [H: NoValue = NoValue |- _] => inversion H; subst; clear H
    | [H: realizeState _ _ _ |- _] => inversion H; subst; clear H
    | [H: basicState _ _ _ |- _] => inversion H; subst; clear H
    (*| [H: absEval (AbsQVar _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsHeapRef _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsPlus _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsMinus _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsTimes _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsImply _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsNot _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsAnd _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsOr _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsIte _ _ _) _ _ |- _] => inversion H; subst; clear H
    | [H: absEval (AbsEqual _ _) _ (NatValue 0) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsEqual _ _) _ (NatValue 1) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsLess _ _) _ (NatValue 0) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsLess _ _) _ (NatValue 1) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsRMember _ _ _ _ _) _ (NatValue 0) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsRInclude _ _ _ _ _) _ (NatValue 0) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsRMember _ _ _ _ _) _ (NatValue 1) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsRInclude _ _ _ _ _) _ (NatValue 1) |- _] => inversion H; subst; clear H
    | [H: absEval (AbsEqual _ _) _ (NatValue 1) |- _] => inversion H; subst; clear H*)
    | [H: context [instantiateState _ _] |- _ ] => simpl in H
    | [H: exists _, _ |- _] => inversion H; subst; clear H
    | [H: context [(match env_p (id->option nat) ?X ?Y ?Z with | Some _ => _ | None => _ end)] |- _] => let xx:=fresh in remember (env_p (id->option nat) X Y Z) as xx; destruct xx; simpl in H
    (*| [H: context [btnat _] |- _] => unfold btnat in H*)
    | [H: context [basicEval _ _] |- _] => unfold basicEval in H
    (*| [H: context [AbsEqualId] |- _] => unfold AbsEqualId in H*)
    | [H: context [absCanEvalList _ _] |- _] => simpl in H
    | [H: Some _ = Some _ |- _] => inversion H; subst; clear H
    end.

Ltac propagate :=
           match goal with
           | [H: fst ?s _ = Some _, C:concreteCompose ?s _ _ |- _] => erewrite composeEnvPropagate1 in H;[idtac|apply C]
           | [H: fst ?s _ = Some _, C:concreteCompose _ ?s _ |- _] => erewrite composeEnvPropagate2 in H;[idtac|apply C]
           | [H: Some _ = fst ?s _, C:concreteCompose ?s _ _ |- _] => erewrite composeEnvPropagate1 in H;[idtac|apply C]
           | [H: Some _ = fst ?s _, C:concreteCompose _ ?s _ |- _] => erewrite composeEnvPropagate2 in H;[idtac|apply C]
           end.

Ltac reduceHypothesis := repeat reduceHyp.

Ltac simplHyp := repeat match goal with
                 | [H: _ = fst (_,_) _ |- _] => simpl in H
                 | [H: fst (_,_) _ = _ |- _] => simpl in H
                 | [H1: ?x ?y = Some _, H2: ?x ?y = Some _ |- _] => rewrite H1 in H2
                 | [H1: Some _ = ?x ?y, H2: Some _ = ?x ?y |- _] => rewrite <- H1 in H2
                 | [H: Some _ = Some _ |- _] => inversion H; subst; clear H
                 end.

Ltac tryEmptyCompose := apply emptyConcreteCompose || idtac.

Ltac simplifyEval :=
     match goal with
     | [H: Some _ = ?e ?v |- context [?e ?v]] => rewrite <- H;simpl;try simplifyEval
     | [H: ?e ?v = Some _ |- context [?e ?v]] => rewrite H;simpl;try simplifyEval
     | [ |- context [env_p _ _ _ _]] => unfold env_p;simpl;try simplifyEval
     | [ |- context [env_p _ _ _]] => unfold env_p;simpl;try simplifyEval
     (*| [ |- context [btnat _]] => unfold btnat*)
     | [ |- _ = _] => reflexivity
     (*| [ |- context [AbsEqualId] ] => unfold AbsEqualId; simpl; try simplifyEval
     | [ |- context [AbsEqualId] ] => unfold AbsOrId; simpl; try simplifyEval
     | [ |- context [AbsEqualId] ] => unfold AbsRMemberId; simpl; try simplifyEval*)
     | [ |- context [basicEval (Id _) _] ] => unfold basicEval; simpl; try simplifyEval
     end.

Ltac decomposeR :=
    match goal with
    | [ |- Tree 0 _ _ _ empty_heap ] => eapply TreeBase;[omega | unfold empty_heap; reflexivity ]
    end.

Ltac decomposeBasicState :=
    match goal with
    (*| [ |- context [Id 2] ] => try decomposeBasicState*)
    | [ |- basicState (AbsTreeId) (NatValue _::_::NatValue _::_) _] => eapply BStateTree; [try decomposeR | simplifyStripNatValues]
    | [ |- basicState (AbsPredicateId) _ _] => eapply BTStatePredicate;[ omega | intros; simpl; reflexivity ]
    | [ |- _ ] => simpl; simplifyEval; try decomposeBasicState
    end.

Ltac decomposeTheState :=
    match goal with
    | [ |- realizeState (AbsStar _ _) _ _] => eapply RSCompose;[try decomposeTheState | try decomposeTheState | tryEmptyCompose]
    | [ |- realizeState AbsEmpty _ _] => (eapply RSEmpty;simpl;unfold empty_heap;reflexivity)
    | [ |- realizeState (AbsLeaf _ _) _ _] => eapply RSR;[simpl;reflexivity | idtac]
    | [ |- realizeState (AbsAll _ _) _ _] => eapply RSAll;[simpl;reflexivity | simpl;reflexivity | simpl | idtac]
    (*| [ |- realizeStateList ((AbsExistsT _)::_) _] => eapply RSLExistsU;eapply ex_intro;simpl
    | [ |- realizeStateList ((AbsExists _ _ _)::_) _] => eapply RSLExists;eapply ex_intro;simpl
    | [ |- realizeStateList ((AbsEach _ _ _)::_) _] => eapply RSLEach;intros;simpl;decomposeTheState
    | [ |- realizeStateList ((AbsAccumulate _ _ _ _ _ _)::_) _ ] => eapply RSLAccumulate;[simpl;reflexivity|simpl;reflexivity|idtac|simpl;reflexivity]
    | [ |- absEval _ _ _] => decomposeEval
    | [ H:?X <>?X |- _] => let H1 := fresh in assert (X = X) as H1; reflexivity; apply H in H; inversion H*)
    end.
