(**********************************************************************************
 * The PEDANTIC (Proof Engine for Deductive Automation using Non-deterministic
 * Traversal of Instruction Code) verification framework
 *
 * Developed by Kenneth Roe
 * For more information, check out www.cs.jhu.edu/~roe
 *
 * Merge.v
 * This file contains many useful definitions for merging two abstract states into
 * a single state implied by both.
 *
 * Key definitions:
 *     MergeStatesTheorem
 *     doMergeStates
 *
 **********************************************************************************)

Require Export SfLib.
Require Export ImpHeap.
Require Export AbsState.
Require Export PickElement.
Require Export AbsExecute.
Require Export Coq.Logic.FunctionalExtensionality.
Require Export Fold.

Fixpoint merge_equiv {ev} {eq} {f} (a : list (@absExp ev eq f)) (t : (@absExp ev eq f)) (b : list (list (@absExp ev eq f))) :=
    match b with
    | nil => ((t::a)::nil)
    | (x::y) => if mem_absExp t x then ((a++x)::y) else x::(merge_equiv a t y)
    end.

Fixpoint add_equiv {ev} {eq} {f} (equiv : (list (list (@absExp ev eq f)))) (t1 : (@absExp ev eq f)) (t2 : (@absExp ev eq f)) :=
    match equiv with
    | (a::b) => if mem_absExp t1 a then
                    if mem_absExp t2 a then (a::b) else merge_equiv a t2 b
                else
                    if mem_absExp t2 a then merge_equiv a t1 b else a::(add_equiv b t1 t2)
    | nil => (t1::t2::nil)::nil
    end.

Inductive build_equivalents {ev} {eq} {f} {t} {ac} : @absState ev eq f t ac -> (list (list (@absExp ev eq f))) -> Prop :=
    | BEAdd : forall s s' l r e e',
            spickElement s ([l====r]) s' ->
            build_equivalents s' e ->
            e' = add_equiv e l r ->
            build_equivalents s e'
    | BEDone : forall s, build_equivalents s nil.

Ltac solveBuildEquivalents := solve [eapply BEAdd;[solveSPickElement|solveBuildEquivalents|(simpl;reflexivity)]||eapply BEDone].

Fixpoint map_var (m : list (nat * nat)) (v : nat) : option nat :=
   match m with
   | nil => None
   | ((a,x)::r) => if beq_nat x v then Some a else map_var r v
   end.

Fixpoint fold_list {EV} {EQ} {T} (l : list (option (@absExp EV EQ T))) :=
   match l with
   | nil => Some nil
   | (f::r) => match f,fold_list r with
               | Some x,Some y => Some (x::y)
               | _,_ => None
               end
   end.

Fixpoint mapAbsQVar {EV} {EQ} {T} (m : list (nat * nat)) (e : @absExp EV EQ T) : option (@absExp EV EQ T) :=
   match e with
   | AbsConstVal v => Some (AbsConstVal v)
   | AbsVar vv => Some (AbsVar vv)
   | AbsQVar v => match map_var m v with | None => None | Some x => Some (AbsQVar x) end
   | AbsFun i el => match fold_list (map (mapAbsQVar m) el) with
                    | None => None
                    | Some l => Some (AbsFun i l)
                    end 
   end.

Fixpoint mapAbsQVarState {EV} {EQ} {F} {T} {AC} (m : list (nat * nat)) (s : @absState EV EQ F T AC) : option (@absState EV EQ F T AC) :=
   match s with
    | AbsStar s1 s2 => match mapAbsQVarState m s1,mapAbsQVarState m s2 with
                       | Some a, Some b => Some (AbsStar a b)
                       | _, _ => None
                       end
    | AbsExists e s => match mapAbsQVar m e, mapAbsQVarState m s with
                       | Some a, Some b => Some (AbsExists a b)
                       | _, _ => None
                       end
    | AbsExistsT s => match mapAbsQVarState m s with
                      | Some x => Some (AbsExistsT x)
                      | None => None
                      end
    | AbsAll e s => match mapAbsQVar m e, mapAbsQVarState m s with
                    | Some a, Some b => Some (AbsAll a b)
                    | _, _ => None
                    end
    | AbsEach e s => match mapAbsQVar m e, mapAbsQVarState m s with
                     | Some a, Some b => Some (AbsEach a b)
                     | _, _ => None
                     end
    | AbsEmpty => Some AbsEmpty
    | AbsAccumulate a b c d e f => None
    | AbsLeaf i el => match fold_list (map (mapAbsQVar m) el) with
                      | None => None
                      | Some l => Some (AbsLeaf i l)
                      end
   end.


(*
 * This inductive definition contains the meat of the merging environment.
 * Note the use of pick2RsNi.  There is a subtle issue that the merge operation
 * also has the secondary task of finding an appropriate mapping between
 * bound variables in the two states.  pick2RsNi requires that any quantified
 * variables being matched up in the first parameter of a TREE or cell construct,
 * or the only term in a predicate construct, that the mappings already be present.
 * This forces a top down match of the states.  For example, merging the following
 * two states:
 *     X++++#0 |-> v(0) *
 *     X++++#1 |-> v(1) *
 *     TREE(v(0),v(4),#2,#0::nil)
 * and
 *     X++++#0 |-> v(3) *
 *     X++++#1 |-> v(4) *
 *     TREE(v(0),v(5),#2,#0::nil)
 * will produce:
 *     X++++#0 |-> v(0) *
 *     X++++#1 |-> v(1) *
 *     TREE(v(0),v(4),#2,#0::nil)
 *
 * However matching,
 *     v(1)++++#0 |-> v(0) *
 *     v(1)++++#1 |-> v(1) *
 *     TREE(v(0),v(4),#2,#0::nil)
 * and
 *     v(2)++++#0 |-> v(3) *
 *     v(2)++++#1 |-> v(4) *
 *     TREE(v(0),v(5),#2,#0::nil)
 * will not match unless the input mapping alread maps v(1) to v(2).
 *
 * The parameters are as follows:
 *
 *    #1 : absState - input of the first of the two states being merged
 *                    (top AbsExistsT quantifiers stripped)
 *    #2 : absState - input of the second of the two states being merged
 *                    (top AbsExistsT quantifiers stripped)
 *    #3 : nat - # of quantified variables in first state
 *    #4 : nat - # of quantified variables in second state
 *    #5 : (list (nat * nat)) - input mapping between bound variables
 *    #6 : list (list absExp) - equivalent expressions in first state (for pick2RsNi)
 *    #7 : list (list absExp) - equivalent expressions in second state (for pick2RsNi)
 *    #8 : absState - output of the merged state (with AbsExistsT's stripped off)
 *    #9 : list (nat * nat) - output mapping between bound variables
 *                            (mappings added by pick2RsNi)
 *)
Inductive doMergeStates {ev} {eq} {f} {t} {ac} : @absState ev eq f t ac -> 
                                                 @absState ev eq f t ac ->
                                                 nat -> nat -> (list (nat * nat)) ->
                                                 (list (list (@absExp ev eq f))) ->
                                                 (list (list (@absExp ev eq f))) ->
                                                 @absState ev eq f t ac ->
                                                 (list (nat *nat)) -> Prop :=
     (* Pair two arbitrary predicates by 'oring' them together in the merged state *)
   | DMOrPredicates1: forall a b a' b' m m' t1 t2 equiv1 equiv2 p1 p2 p2' res,
       spickElement a ([p1]) a' ->
       spickElement b ([p2]) b' ->
       Some p2' = @mapAbsQVar ev eq f m p2 ->
       doMergeStates a' b' t1 t2 m equiv1 equiv2 res m' ->
       doMergeStates a b t1 t2 m equiv1 equiv2 (([p1 \\// p2']) ** res) m'
      (* Pair two predicates where the one from the second state implies the one from the first state *)
   | DMImplyPredicates1: forall a b a' b' m m' t1 t2 equiv1 equiv2 p1 p2 p2' res,
       spickElement a ([p1]) a' ->
       spickElement b ([p2]) b' ->
       Some p2' = @mapAbsQVar ev eq f m p2 ->
       (forall e b, @absEval ev eq f e b p2' = NatValue 0 -> @absEval ev eq f e b p1 = NatValue 0) ->
       doMergeStates a' b' t1 t2 m equiv1 equiv2 res m' ->
       doMergeStates a b t1 t2 m equiv1 equiv2 (([p2']) ** res) m'
      (* Pair two predicates where the one from the first state implies the one from the second state *)
   | DMImplyPredicates2: forall a b a' b' m m' t1 t2 equiv1 equiv2 p1 p2 p2' res,
       spickElement a ([p1]) a' ->
       spickElement b ([p2]) b' ->
       Some p2' = @mapAbsQVar ev eq f m p2 ->
       (forall e b, @absEval ev eq f e b p1 = NatValue 0 -> @absEval ev eq f e b p2' = NatValue 0) ->
       doMergeStates a' b' t1 t2 m equiv1 equiv2 res m' ->
       doMergeStates a b t1 t2 m equiv1 equiv2 (([p1]) ** res) m'
     (* Finish the pairing process--can only be done when only predicates are left *)
   | DMFinish : forall a b t1 t2 equiv1 equiv2 m m',
       allPredicates a ->
       allPredicates b ->
       doMergeStates a b t1 t2 m equiv1 equiv2 AbsEmpty m'
     (* Remove identical RFun constructs from the two abstract states *)
   | DMRemoveR : forall a b equiv1 equiv2 t1 t2 h1 h2 m' m'' a' b' m res,
       pick2RsNi a b m t1 t2 equiv1 equiv2 h1 h2 a' b' m' ->
       doMergeStates a' b' t1 t2 m' equiv1 equiv2 res m'' ->
       doMergeStates a b t1 t2 m equiv1 equiv2 (h1 ** res) m''.

(*
 * This is the theorem that sets up state merging through pairing of the components
 *)
Theorem MergeStatesTheorem {ev} {eq} {f} {t} {ac} : forall s1 s2 t1 t2 s1' s2' m sm sm',
    (s1',t1) = @remove_top_existentials ev eq f t ac s1 ->
    (s2',t2) = @remove_top_existentials ev eq f t ac s2 ->
    (*build_equivalents s1' equiv1 ->
    build_equivalents s2' equiv2 ->*)
    doMergeStates s1' s2' t1 t2 nil nil nil sm m ->
    sm' = replaceRoot s1 sm ->
    mergeStates s1 s2 sm'.
Proof. admit. Qed.

(*
 * Tactics for automating the merge process
 *)
Ltac doMergeStates :=
        (eapply DMRemoveR; [solvePick2RsNi | doMergeStates]) ||
        (idtac).

Ltac startMerge :=
    eapply MergeStatesTheorem;[
        (simpl; reflexivity) |
        (simpl; reflexivity) |
        doMergeStates |
        (simpl; reflexivity)].

Ltac finishMerge := eapply DMFinish;[solveAllPredicates | solveAllPredicates].

(* **************************************************************************
 *
 * Finally, we have auxiliary theorems for performing a fold before merging.
 *
 ****************************************************************************)

Theorem foldLeft : forall ev eq f t ac s s' sx m,
    @foldHeap ev eq f t ac s s' ->
    mergeStates s' sx m ->
    mergeStates s sx m.
Proof. admit. Qed.

Theorem foldRight : forall ev eq f t ac s s' sx m,
    @foldHeap ev eq f t ac s s' ->
    mergeStates sx s' m ->
    mergeStates sx s m.
Proof. admit. Qed.

Theorem foldAllLeft : forall ev eq f t ac s s' sx m,
    @foldAll ev eq f t ac s s' ->
    mergeStates s' sx m ->
    mergeStates s sx m.
Proof. admit. Qed.

Theorem foldAllRight : forall ev eq f t ac s s' sx m,
    @foldAll ev eq f t ac s s' ->
    mergeStates sx s' m ->
    mergeStates sx s m.
Proof. admit. Qed.
